#ifndef BBCmd_BBCMD_H
#define BBCmd_BBCMD_H

#ifdef __cplusplus
extern "C" {
#endif

/**
 * This is the public header file of BBCmd.
 * The modules that just use BBCmd API should include this.
 */




#define BBCmd_RELEASE_LEVEL_ALPHA 0xA
#define BBCmd_RELEASE_LEVEL_BETA 0xB
#define BBCmd_RELEASE_LEVEL_GAMMA 0xC
#define BBCmd_RELEASE_LEVEL_FINAL 0xF

#define BBCmd_MAJOR_VERSION 0
#define BBCmd_MINOR_VERSION 1
#define BBCmd_MICRO_VERSION 6
#define BBCmd_RELEASE_LEVEL BBCmd_RELEASE_LEVEL_FINAL

#define _BBCmd_VERSION(a,b,c,d) (((a)<<24) | ((b)<<16) | ((c)<<8) | ((d)<<4))
#define BBCmd_VERSION _BBCmd_VERSION( BBCmd_MAJOR_VERSION,	\
                                      BBCmd_MINOR_VERSION,	\
                                      BBCmd_MICRO_VERSION,	\
                                      BBCmd_RELEASE_LEVEL )




/**
 * For debugging.
 * Define BBCmd_IS_DEBUG_VERSION to enable these feature.
 */
#ifdef BBCmd_IS_DEBUG_VERSION

/* Just counting the number of allocation. */
#include "_malloc.h"

#endif




/* To declare a public function, data or type. */
#define BBCmdAPI_FUNC(type) __declspec(dllexport) extern type
#define BBCmdAPI_DATA(type) __declspec(dllexport) extern type
#define BBCmdAPI_TYPE(type) type




/* To represent a `command'. */
BBCmdAPI_TYPE(typedef struct _BBCmdCommand) BBCmdCommand;

/* To represent the descriptor of a `command'.
 * A descriptor contains the name, document and functions of the command
 * that is represented by the descriptor.  */
BBCmdAPI_TYPE(typedef struct _BBCmdCommandDescriptor) BBCmdCommandDescriptor;

/* To represent a `compiler' that compile `command' from other object. */
BBCmdAPI_TYPE(typedef struct _BBCmdCompiler) BBCmdCompiler;




/**
 * Make a new instance of compiler and return it
 * (or NULL for failure that may be caused by running out of memory).
 *
 * BBCmdCompiler_New makes a new instance of compiler with default
 * command definition that might be useful to operate BB4W.
 *
 * BBCmdCompiler_NewVanilla makes a new instance of compiler with
 * no command definition.  The instance might be useful to make
 * your original compiler that doesn't have to have default command definition.
 */
BBCmdAPI_FUNC(BBCmdCompiler*) BBCmdCompiler_New(void);
BBCmdAPI_FUNC(BBCmdCompiler*) BBCmdCompiler_NewVanilla(void);

/* Delete a given instance of compiler. */
BBCmdAPI_FUNC(void) BBCmdCompiler_Del(BBCmdCompiler*);

/**
 * Compile command from a given string by a given compiler.
 * Return a new instance of command.
 *
 * If this function fail, then this will return NULL
 * and store the information about the error
 * to *error_line, *error_column and *error_message.
 *
 * See also the source code of BBCmdCompiler_CompileFromString
 * for the format of script.
 */
BBCmdAPI_FUNC(BBCmdCommand*)
BBCmdCompiler_CompileFromString( const BBCmdCompiler*,
                                 const char* script,
                                 int* error_line,
                                 int* error_column,
                                 const char** error_message );

/**
 * Add a new command definition to a specified compiler.
 * Return true (nonzero) for success or false (zero) for failure.
 *
 * BUGS: BBCmdCompiler assums that a given descriptor exists
 *       while running process.
 *
 * BUGS: Adding a command definition
 *       that has the same name as existing commands in the compiler
 *       will be hidden by the commands.
 *
 *       See the source code of
 *       _BBCmdCompiler_FindDescriptor and BBCmdCompiler_AddNewCommand
 *       for this mechanism.
 */
BBCmdAPI_FUNC(int)
BBCmdCompiler_AddNewCommand(BBCmdCompiler*, const BBCmdCommandDescriptor*);

/**
 * Return available command (descriptor) for a specified compiler
 * as the pointer to the array of the descriptors.
 * The number of the descriptors will be stored `*n' if `n' is not NULL.
 *
 * Note: If `*n' is zero, the returning value will be NULL.
 */
BBCmdAPI_FUNC(const BBCmdCommandDescriptor* const*)
BBCmdCompiler_AvailableCommand(const BBCmdCompiler*, int* n);




/**
 * Delete a given instance of command.
 * Passing NULL is just ignored.
 */
BBCmdAPI_FUNC(void) BBCmdCommand_Del(BBCmdCommand*);

/**
 * Execute a given instance of command and return its result.
 * _BBCmdCommand_Execute can return negative value.
 */
BBCmdAPI_FUNC(int) BBCmdCommand_Execute(const BBCmdCommand*);
BBCmdAPI_FUNC(int) _BBCmdCommand_Execute(const BBCmdCommand*);




/* Get the specified information of a given command descriptor. */
BBCmdAPI_FUNC(const char*)
BBCmdCommandDescriptor_NameOf(const BBCmdCommandDescriptor*);
BBCmdAPI_FUNC(const char* const*)
BBCmdCommandDescriptor_AliasOf(const BBCmdCommandDescriptor*);
BBCmdAPI_FUNC(const char*)
BBCmdCommandDescriptor_DocOf(const BBCmdCommandDescriptor*);




#define WIN32_LEAN_AND_MEAN
#include <windows.h>

/* Return the window handle of the core */
BBCmdAPI_FUNC(HWND) BBCmd_FindCore(void);

/* Return the window handle of the proxy */
BBCmdAPI_FUNC(HWND) BBCmd_FindProxy(void);

/* Return wether this process is a plugin of BB4W */
BBCmdAPI_FUNC(int) BBCmd_AmIPlugin(void);

/**
 * Return the current workspace number (0, 1, ... n-1).
 *
 * If this process is not a plugin of BB4W
 * or this function failed by other reason,
 * this function will return negative value.
 */
BBCmdAPI_FUNC(int) BBCmd_GetCurrentWorkspaceNumber(void);

/* Return the number of the elements of a given array */
#define BBCmd_arraysizeof(a) (sizeof(a) / sizeof((a)[0]))

/* Return whether the library is compatible with the caller. */
BBCmdAPI_FUNC(int) BBCmd_IsCompatible(int version, int hint);

	/* hint value for BBCmd_IsCompatible */
#define BBCmd_ICH_JUST_USE (0<<0)
#define BBCmd_ICH_KNOW_INTERNAL (1<<0)

	/* result value of BBCmd_IsCompatible */
#define BBCmd_IC_COMPATIBLE 0
#define BBCmd_IC_TOO_OLD_DLL 1
#define BBCmd_IC_INCOMPATIBLE_API 2
#define BBCmd_IC_INCOMPATIBLE_INTERNAL 3




#ifdef __cplusplus
}
#endif

#endif
