/* Implementation of `mouse_move' and other mouse-related commands */

#include <assert.h>
#include <stdlib.h>

#define WIN32_LEAN_AND_MEAN
#include <windows.h>

#include "Mouse_p.h"




/* -------- mouse_move  -------- */

static const char*
_BBCmdMouseMoveCommand_initialize( BBCmdCommand* _self,
                                   int ac, const char* const* av,
                                   const BBCmdCompiler* compiler )
{
	BBCmdMouseMoveCommand* self;

	assert(_self->cmd_descriptor == BBCmdMouseMoveCommandDescriptor);
	self = (BBCmdMouseMoveCommand*)_self;

	self->mm_x = NULL;
	self->mm_y = NULL;

	if (ac - 1 != 2) return "Incorrect number of arguments";

	self->mm_x = strdup(av[1]);
	self->mm_y = strdup(av[2]);
	if (self->mm_x == NULL || self->mm_y == NULL)
		return "Out of memory (_BBCmdMouseMoveCommand_initialize)";

	return NULL;
}


static void
_BBCmdMouseMoveCommand_del(BBCmdCommand* _self)
{
	BBCmdMouseMoveCommand* self;

	assert(_self->cmd_descriptor == BBCmdMouseMoveCommandDescriptor);
	self = (BBCmdMouseMoveCommand*)_self;

	free(self->mm_x);
	free(self->mm_y);
}


static int
_BBCmdMouseMoveCommand_execute(const BBCmdCommand* _self)
{
	BBCmdMouseMoveCommand* self;
	int x;
	int y;
	POINT pt;

	assert(_self->cmd_descriptor == BBCmdMouseMoveCommandDescriptor);
	self = (BBCmdMouseMoveCommand*)_self;

	if (    self->mm_x[0] == '+' || self->mm_x[0] == '-'
	     || self->mm_y[0] == '+' || self->mm_y[0] == '-'  )
	{
		GetCursorPos(&pt);
	}

	if      (self->mm_x[0] == '+') x = pt.x + atoi(self->mm_x + 1);
	else if (self->mm_x[0] == '-') x = pt.x - atoi(self->mm_x + 1);
	else                           x =        atoi(self->mm_x);

	if      (self->mm_y[0] == '+') y = pt.y + atoi(self->mm_y + 1);
	else if (self->mm_y[0] == '-') y = pt.y - atoi(self->mm_y + 1);
	else                           y =        atoi(self->mm_y);

	SetCursorPos(x, y);

	return BBCmd_RC_TRUE;
}




const BBCmdCommandDescriptor _BBCmdMouseMoveCommandDescriptor = {
	"mouse_move",
	NULL,
	"(x, y) --> true\n"
		"Move the cursor to the specified coordinates.\n"
		"Prefix `+' or `-' to specify relative coordinate.",
	sizeof(BBCmdMouseMoveCommand),
	_BBCmdMouseMoveCommand_initialize,
	_BBCmdMouseMoveCommand_del,
	_BBCmdMouseMoveCommand_execute
};




/* -------- mouse_wheel  -------- */

static const char*
_BBCmdMouseWheelCommand_initialize( BBCmdCommand* _self,
                                    int ac, const char* const* av,
                                    const BBCmdCompiler* compiler )
{
	BBCmdMouseWheelCommand* self;

	assert(_self->cmd_descriptor == BBCmdMouseWheelCommandDescriptor);
	self = (BBCmdMouseWheelCommand*)_self;

	if (ac - 1 != 1) return "Incorrect number of arguments";

	self->mw_delta = atoi(av[1]);

	return NULL;
}


static void
_BBCmdMouseWheelCommand_del(BBCmdCommand* _self)
{
	/* nothing to do */
}


static int
_BBCmdMouseWheelCommand_execute(const BBCmdCommand* _self)
{
	BBCmdMouseWheelCommand* self;

	assert(_self->cmd_descriptor == BBCmdMouseWheelCommandDescriptor);
	self = (BBCmdMouseWheelCommand*)_self;

	mouse_event(MOUSEEVENTF_WHEEL, 0, 0, self->mw_delta, 0);

	return BBCmd_RC_TRUE;
}




const BBCmdCommandDescriptor _BBCmdMouseWheelCommandDescriptor = {
	"mouse_wheel",
	NULL,
	"(delta)\n"
		"Rotate the mouse wheel.\n"
		"The amount of wheel movement is specified by `delta'.\n"
		"A positive value will rotate forward,\n"
		"and a nevative value will rotate backward.\n"
		"One wheel click is defined as 120.",
	sizeof(BBCmdMouseWheelCommand),
	_BBCmdMouseWheelCommand_initialize,
	_BBCmdMouseWheelCommand_del,
	_BBCmdMouseWheelCommand_execute
};




/* -------- mouse_click  -------- */

#ifndef XBUTTON1
#define XBUTTON1 0x0001
#endif
#ifndef XBUTTON2
#define XBUTTON2 0x0002
#endif

#ifndef MOUSEEVENTF_XDOWN
#define MOUSEEVENTF_XDOWN 0x0080
#endif
#ifndef MOUSEEVENTF_XUP
#define MOUSEEVENTF_XUP 0x0100
#endif

#define MODIFIER_SHIFT (1<<(0+8))
#define MODIFIER_CTRL (1<<(1+8))
#define MODIFIER_ALT (1<<(2+8))
#define MODIFIER_WIN (1<<(3+8))
#define MODIFIER_UP (1<<(4+8))

#define BTN_LEFT (1<<0)
#define BTN_RIGHT (1<<1)
#define BTN_MIDDLE (1<<2)
#define BTN_X1 (1<<3)
#define BTN_X2 (1<<4)

static const char*
_BBCmdMouseClickCommand_initialize( BBCmdCommand* _self,
                                    int ac, const char* const* av,
                                    const BBCmdCompiler* compiler )
{
	BBCmdMouseClickCommand* self;
	int i;

	assert(_self->cmd_descriptor == BBCmdMouseClickCommandDescriptor);
	self = (BBCmdMouseClickCommand*)_self;

	self->mc_data = NULL;

	if (ac - 1 < 1) return "Incorrect number of arguments";

	self->mc_data = (int*)malloc(sizeof(self->mc_data[0]) * (ac-1 + 1));
	if (self->mc_data == NULL) return "Out of memory";

	for (i = 0; i < ac-1; i++) {
		const char* p;

		p = av[1+i];

		self->mc_data[i] = 0;

		while (p[0] != '\0' && p[1] != '\0') {
			switch (p[0]) {
			default: return "Syntax error - invalid modifier";
			case 'S': self->mc_data[i] |= MODIFIER_SHIFT; break;
			case 'C': self->mc_data[i] |= MODIFIER_CTRL; break;
			case 'A':
			case 'M': self->mc_data[i] |= MODIFIER_ALT; break;
			case 'W': self->mc_data[i] |= MODIFIER_WIN; break;
			}

			p += 2;
		}

		if (p[0] == '\0') return "Syntax error";

		switch (p[0]) {
		default: return "Syntax error - invalid button";
		case 'l': self->mc_data[i] |= MODIFIER_UP;  /* FALLTHRU */
		case 'L': self->mc_data[i] |= BTN_LEFT; break;
		case 'r': self->mc_data[i] |= MODIFIER_UP;  /* FALLTHRU */
		case 'R': self->mc_data[i] |= BTN_RIGHT; break;
		case 'm': self->mc_data[i] |= MODIFIER_UP;  /* FALLTHRU */
		case 'M': self->mc_data[i] |= BTN_MIDDLE; break;
		case 'x': self->mc_data[i] |= MODIFIER_UP;  /* FALLTHRU */
		case 'X': self->mc_data[i] |= BTN_X1; break;
		case 'y': self->mc_data[i] |= MODIFIER_UP;  /* FALLTHRU */
		case 'Y': self->mc_data[i] |= BTN_X2; break;
		}
	}

	self->mc_data[i] = 0;

	return NULL;
}


static void
_BBCmdMouseClickCommand_del(BBCmdCommand* _self)
{
	BBCmdMouseClickCommand* self;

	assert(_self->cmd_descriptor == BBCmdMouseClickCommandDescriptor);
	self = (BBCmdMouseClickCommand*)_self;

	free(self->mc_data);
}


static BYTE
mod_as_vk(int mod)
{
	switch (mod) {
	default: assert(0);
	case MODIFIER_SHIFT: return VK_SHIFT;
	case MODIFIER_CTRL: return VK_CONTROL;
	case MODIFIER_ALT: return VK_MENU;
	case MODIFIER_WIN: return VK_LWIN;
	}
}

static void
btn_as_mouse_event_args(int btn, int* flag, int* data)
{
	*data = 0;

	switch (btn & 0x00FF) {
	default: assert(0);
	case BTN_LEFT:
		*flag = ( (btn & MODIFIER_UP) ? MOUSEEVENTF_LEFTUP
		                              : MOUSEEVENTF_LEFTDOWN );
		break;
	case BTN_RIGHT:
		*flag = ( (btn & MODIFIER_UP) ? MOUSEEVENTF_RIGHTUP
		                              : MOUSEEVENTF_RIGHTDOWN );
		break;
	case BTN_MIDDLE:
		*flag = ( (btn & MODIFIER_UP) ? MOUSEEVENTF_MIDDLEUP
		                              : MOUSEEVENTF_MIDDLEDOWN );
		break;
	case BTN_X1:
	case BTN_X2:
		*flag = ( (btn & MODIFIER_UP) ? MOUSEEVENTF_XUP
		                              : MOUSEEVENTF_XDOWN );
		*data = ((btn & BTN_X1) ? XBUTTON1 : XBUTTON2);
		break;
	}

	if (GetSystemMetrics(SM_SWAPBUTTON)) {
		switch (*flag) {
		default: break;
		case MOUSEEVENTF_LEFTUP: *flag=MOUSEEVENTF_RIGHTUP; break;
		case MOUSEEVENTF_LEFTDOWN: *flag=MOUSEEVENTF_RIGHTDOWN; break;
		case MOUSEEVENTF_RIGHTUP: *flag=MOUSEEVENTF_LEFTUP; break;
		case MOUSEEVENTF_RIGHTDOWN: *flag=MOUSEEVENTF_LEFTDOWN; break;
		}
	}
}

static int
_BBCmdMouseClickCommand_execute(const BBCmdCommand* _self)
{
	const int MODIFIERS[] = {
		MODIFIER_SHIFT,
		MODIFIER_CTRL,
		MODIFIER_ALT,
		MODIFIER_WIN,
		0
	};
	const int* m;
	BBCmdMouseClickCommand* self;
	int i;
	int flag;
	int data;

	assert(_self->cmd_descriptor == BBCmdMouseClickCommandDescriptor);
	self = (BBCmdMouseClickCommand*)_self;

	for (i = 0; self->mc_data[i] != 0; i++) {
		for (m = MODIFIERS; *m != 0; m++) {
		  if (self->mc_data[i] & *m)
		    keybd_event(mod_as_vk(*m), 0, 0, 0);
		  else
		    keybd_event(mod_as_vk(*m), 0, KEYEVENTF_KEYUP, 0);
		}

		btn_as_mouse_event_args(self->mc_data[i], &flag, &data);
		mouse_event(flag, 0, 0, data, 0);

		for (m = MODIFIERS; *m != 0; m++) {
		  if (self->mc_data[i] & *m)
		    keybd_event(mod_as_vk(*m), 0, KEYEVENTF_KEYUP, 0);
		}
	}

	return BBCmd_RC_TRUE;
}




const BBCmdCommandDescriptor _BBCmdMouseClickCommandDescriptor = {
	"mouse_click",
	NULL,
	"(click_spec...)\n"
		"Emulate to click a mouse button.\n"
		"`click_spec' can be one of the following value:\n"
		"\n"
		"  L/l  Press/release left button\n"
		"  R/r  ... right\n"
		"  M/m  ... middle\n"
		"  X/x  ... X1\n"
		"  Y/y  ... X2\n"
		"\n"
		"You can also specify some modifiers:\n"
		"\n"
		"  S    Shift\n"
		"  C    Control\n"
		"  A    Alt\n"
		"  M    same as A\n"
		"  W    Win\n"
		"\n"
		"For example,\n"
		"`mouse_click('L', 'l')' will emulate left-click,\n"
		"`mouse_click('C-R', 'C-r')' will emulate ctrl-right-click.",
	sizeof(BBCmdMouseClickCommand),
	_BBCmdMouseClickCommand_initialize,
	_BBCmdMouseClickCommand_del,
	_BBCmdMouseClickCommand_execute
};




/* __END__ */
